using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;

namespace SharpRSS {
	public partial class Form1 : Form {
		public Form1()
		{
			InitializeComponent();
			Text = Program.ApplicationName;
		}

		class FeedSite {
			public FeedSite()
			{
			}
			public string URL;
			public string Title;
			public string Description;
			//public RSS.Channel channel;
		}
		List<FeedSite> _site_list = new List<FeedSite>();

		public static string EscapeHTML(string text)
		{
			if (text == null) {
				return "";
			}
			StringBuilder sb = new StringBuilder();
			foreach (char c in text) {
				switch (c) {
				case '<':
					sb.Append("&lt;");
					break;
				case '>':
					sb.Append("&gt;");
					break;
				case '&':
					sb.Append("&amp;");
					break;
				case '"':
					sb.Append("&quot;");
					break;
				default:
					sb.Append(c);
					break;
				}
			}
			return sb.ToString();
		}

		void SaveFeedList()
		{
			try {
				string path = Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData);
				path = System.IO.Path.Combine(path, "SharpRSS");
				System.IO.Directory.CreateDirectory(path);
				path = System.IO.Path.Combine(path, "feeds.xml");
				using (FileStream s = File.Open(path, FileMode.Create)) {
					StreamWriter w = new StreamWriter(s);
					w.WriteLine("<?xml version='1.0' encoding='utf-8'?>");
					w.WriteLine("<rss version='2.0'>");
					w.WriteLine("  <channel>");
					foreach (FeedSite site in _site_list) {
						w.WriteLine("    <item>");
						w.WriteLine("      <title>" + EscapeHTML(site.Title) + "</title>");
						w.WriteLine("      <link>" + EscapeHTML(site.URL) + "</link>");
						w.WriteLine("      <description>" + EscapeHTML(site.Description) + "</description>");
						w.WriteLine("    </item> ");
					}
					w.WriteLine("  </channel>");
					w.WriteLine("</rss>");
					w.Flush();
					s.Close();
				}
			} catch (Exception e) {
				MessageBox.Show(e.Message);
			}
		}

		void LoadFeedList()
		{
			_site_list.Clear();
			try {
				string path = Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData);
				path = System.IO.Path.Combine(path, "SharpRSS");
				System.IO.Directory.CreateDirectory(path);
				path = System.IO.Path.Combine(path, "feeds.xml");
				RSS.Parser parser = new RSS.Parser();
				RSS.Channel channel;
				List<RSS.Item> items;
				if (!parser.Parse(path, out channel, out items)) {
					throw new Exception(parser.Message);
				}
				foreach (RSS.Item item in items) {
					FeedSite site = new FeedSite();
					site.URL = item.Link;
					site.Title = item.Title;
					site.Description = item.Description;
					_site_list.Add(site);
				}
			} catch (Exception x) {
				MessageBox.Show(x.Message);
			}
		}

		static string FormatDateTime(DateTime dt)
		{
			StringBuilder sb = new StringBuilder();
			sb.Append(dt.Year.ToString("D4"));
			sb.Append("-");
			sb.Append(dt.Month.ToString("D2"));
			sb.Append("-");
			sb.Append(dt.Day.ToString("D2"));
			sb.Append(" ");
			sb.Append(dt.Hour.ToString("D2"));
			sb.Append(":");
			sb.Append(dt.Minute.ToString("D2"));
			sb.Append(":");
			sb.Append(dt.Second.ToString("D2"));
			return sb.ToString();
		}

		TreeNode _uncategorized_node;

		private void UpdateTreeView()
		{
			LoadFeedList();
			treeview.Nodes.Clear();

			_uncategorized_node = treeview.Nodes.Add("");

			foreach (FeedSite site in _site_list) {
				string title = "";
				if (site.Title != null && site.Title.Length > 0) {
					title = site.Title;
				} else {
					title = site.URL;
				}
				TreeNode node = _uncategorized_node.Nodes.Add(title);
				node.Tag = site;
			}

			if (treeview.SelectedNode == null) {
				listview.Items.Clear();
			}
		}

		static int OrderByNameAscend(RSS.Item a, RSS.Item b)
		{
			return a.Title.CompareTo(b.Title);
		}
		static int OrderByNameDescend(RSS.Item a, RSS.Item b)
		{
			return -a.Title.CompareTo(b.Title);
		}

		static int OrderByDateAscend(RSS.Item a, RSS.Item b)
		{
			return a.Date.CompareTo(b.Date);
		}
		static int OrderByDateDescend(RSS.Item a, RSS.Item b)
		{
			return -a.Date.CompareTo(b.Date);
		}

		int _order_by;

		void UpdateListView(RSS.Channel channel, List<RSS.Item> items)
		{
			listview.BeginUpdate();

			switch (_order_by) {
			case 1:
				items.Sort(OrderByNameAscend);
				break;
			case -1:
				items.Sort(OrderByNameDescend);
				break;
			case 2:
				items.Sort(OrderByDateAscend);
				break;
			case -2:
				items.Sort(OrderByDateDescend);
				break;
			}

			listview.Items.Clear();
			foreach (RSS.Item item in items) {
				string[] arr = new string[2];
				arr[0] = item.Title;
				arr[1] = FormatDateTime(item.Date);
				ListViewItem lvi = new ListViewItem(arr);
				listview.Items.Add(lvi);
				lvi.Tag = item;
			}

			foreach (ColumnHeader h in listview.Columns) {
				h.Width = -2;
			}

			listview.EndUpdate();
		}

		void AddFeedByURL(string url)
		{
			RSS.Parser parser = new RSS.Parser();
			RSS.Channel channel;
			List<RSS.Item> items;
			if (!parser.Parse(url, out channel, out items)) {
				MessageBox.Show(parser.Message, Program.ApplicationName);
				return;
			}
			{
				FeedSite site = new FeedSite();
				site.URL = url;
				site.Title = channel.Title;
				site.Description = channel.Description;
				_site_list.Add(site);
			}

			SaveFeedList();
			UpdateTreeView();
		}

		void NavigateSelectedNode()
		{
			if (treeview.SelectedNode == null) {
				return;
			}

			FeedSite site = treeview.SelectedNode.Tag as FeedSite;

			if (site == null) {
				return;
			}

			Capture = true;
			System.Windows.Forms.Cursor.Current = Cursors.WaitCursor;

			{
				string url = site.URL;

				RSS.Parser parser = new RSS.Parser();
				RSS.Channel channel;
				List<RSS.Item> items;
				if (!parser.Parse(url, out channel, out items)) {
					MessageBox.Show(parser.Message, Program.ApplicationName);
					return;
				}

				UpdateListView(channel, items);
			}

			Capture = false;
		}

		private void Form1_Load(object sender, EventArgs e)
		{
			_order_by = -2;
			UpdateTreeView();
		}

		private void vpeBPToolStripMenuItem_Click(object sender, EventArgs e)
		{
			if (treeview.SelectedNode != null) {
				FeedSite site = treeview.SelectedNode.Tag as FeedSite;
				MessageBox.Show(site.URL);
			}
		}

		private void treeview_NodeMouseClick(object sender, TreeNodeMouseClickEventArgs e)
		{
			treeview.SelectedNode = e.Node;
			treeview.ContextMenuStrip = contextMenuStrip1;
		}

		private void Form1_Resize(object sender, EventArgs e)
		{
			foreach (ColumnHeader h in listview.Columns) {
				h.Width = -2;
			}
		}

		bool AcceptTag(string str)
		{
			string s = str.ToLower();
			if (s.IndexOf("script") >= 0) {
				return false;
			}
			if (s.IndexOf("embed") >= 0) {
				return false;
			}
			if (s.IndexOf("object") >= 0) {
				return false;
			}
			if (s.IndexOf("frame") >= 0) {
				return false;
			}
			return true;
		}

		void RegulateHyperText(string text, out string result, out bool usetag)
		{
			StringBuilder sb = new StringBuilder();
			List<string> array = new List<string>();
			bool tag = false;
			foreach (char c in text) {
				if (c == '<') {
					if (!tag) {
						if (sb.Length > 0) {
							array.Add(sb.ToString());
						}
						sb = new StringBuilder();
						sb.Append(c);
						tag = true;
					}
				} else if (c == '>') {
					if (tag) {
						sb.Append(c);
						if (sb.Length > 0) {
							array.Add(sb.ToString());
						}
						sb = new StringBuilder();
						tag = false;
					}
				} else {
					sb.Append(c);
				}
			}
			if (sb.Length > 0) {
				array.Add(sb.ToString());
			}

			usetag = false;

			sb = new StringBuilder();
			foreach (string str in array) {
				if (str.Length > 0) {
					if (str[0] == '<') {
						usetag = true;
						if (AcceptTag(str)) {
							sb.Append(str);
						}
					} else {
						sb.Append(str);
					}
				}
			}

			result = sb.ToString();
		}

		private void listview_SelectedIndexChanged(object sender, EventArgs e)
		{
			if (listview.SelectedItems.Count == 1) {
				RSS.Item item = listview.SelectedItems[0].Tag as RSS.Item;
				if (item != null) {
					label_title.Text = item.Title;
					linklabel.Text = item.Link;
					label_datetime.Text = FormatDateTime(item.Date);
					linklabel.Tag = item;
					//linklabel.Text = item.Link;
					//textbox_description.Text = RegulateHyperText(item.Description);
					string text;
					bool usetag = false;
					RegulateHyperText(item.ContentEncoded + item.Description, out text, out usetag);
					if (!usetag) {
						text = "<html><body><pre>" + text + "</pre></body></html>";
					}
					webBrowser1.DocumentText = text;
				}
			}


		}

		private void \t_Click(object sender, EventArgs e)
		{
			string url = Clipboard.GetText();
			AddFeedByURL(url);
		}

		private void IXToolStripMenuItem_Click(object sender, EventArgs e)
		{
			Close();
		}

		static bool IsValidURL(string url)
		{
			int i = 0;
			if (url.StartsWith("http://")) {
				i = 7;
			} else if (url.StartsWith("https://")) {
				i = 8;
			}
			if (i > 0) {
				i = url.IndexOf('.', i);
				if (i > 0) {
					i = url.IndexOf('/', i);
					if (i > 0) {
						return true;
					}
				}
			}
			return false;
		}

		void Navigate(RSS.Item item)
		{
			if (item != null) {
				string url = item.Link;
				if (IsValidURL(url)) {
					System.Diagnostics.Process.Start(url);
				}
			}
		}

		private void linklabel_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
		{
			Navigate(linklabel.Tag as RSS.Item);
		}

		private void 폜DToolStripMenuItem_Click(object sender, EventArgs e)
		{
			if (treeview.SelectedNode != null) {
				FeedSite site = treeview.SelectedNode.Tag as FeedSite;
				if (site != null) {
					StringBuilder sb = new StringBuilder();
					sb.Append(site.Title);
					sb.Append("\r\n");
					sb.Append(site.Description);
					sb.Append("\r\n");
					sb.Append(site.URL);
					sb.Append("\r\n");
					sb.Append("\r\n");
					sb.Append("RSStB[h폜܂B낵łH");
					if (MessageBox.Show(sb.ToString(), Program.ApplicationName, MessageBoxButtons.OKCancel) != DialogResult.OK) {
						return;
					}
					_site_list.Remove(site);
					SaveFeedList();
					UpdateTreeView();
				}
			}
		}


		protected override bool ProcessKeyPreview(ref Message m)
		{
			if (m.Msg == 0x100 && m.WParam == (IntPtr)0x1b) {
				treeview.Focus();
				return true;
			}
			return base.ProcessKeyPreview(ref m);
		}

		private void o[WAToolStripMenuItem_Click(object sender, EventArgs e)
		{
			AboutForm dlg = new AboutForm();
			dlg.ShowDialog();
		}

		private void listview_KeyPress(object sender, KeyPressEventArgs e)
		{
			if (e.KeyChar == 13) {
				if (listview.SelectedItems.Count > 0) {
					Navigate(listview.SelectedItems[0].Tag as RSS.Item);
				}
			}
		}

		private void listview_DoubleClick(object sender, EventArgs e)
		{
			if (listview.SelectedItems.Count > 0) {
				Navigate(listview.SelectedItems[0].Tag as RSS.Item);
			}
		}

		private void treeview_KeyPress(object sender, KeyPressEventArgs e)
		{
			if (e.KeyChar == 13) {
				NavigateSelectedNode();
			}
		}

		private void treeview_DoubleClick(object sender, EventArgs e)
		{
			NavigateSelectedNode();
		}

		private void listview_ColumnClick(object sender, ColumnClickEventArgs e)
		{
			int n = e.Column + 1;
			if (System.Math.Abs(_order_by) == n) {
				_order_by = -_order_by;
			} else {
				_order_by = n;
			}
			NavigateSelectedNode();
		}

		private void ǉAToolStripMenuItem_Click(object sender, EventArgs e)
		{
			AddFeedForm dlg = new AddFeedForm();
			if (dlg.ShowDialog() != DialogResult.OK) {
				return;
			}
			string url = dlg.URL;
			AddFeedByURL(url);
		}

		private void treeview_DragDrop(object sender, DragEventArgs e)
		{
			string url = e.Data.GetData(typeof(string)).ToString();
			if (IsValidURL(url)) {
				AddFeedByURL(url);
			}
		}

		private void treeview_DragOver(object sender, DragEventArgs e)
		{

		}

		private void treeview_DragEnter(object sender, DragEventArgs e)
		{
			e.Effect = DragDropEffects.All;
		}

	}
}
