
// Yet Another Simple API for XML
// Copyright (C) 2009 Shinichi 'soramimi' Fuchita

using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Text;
using System.IO;

namespace YASAX {
	enum ErrorCode {
		InvalidCharacter,
	}
	class MyException : System.Exception {
		ErrorCode _code;
		public MyException(ErrorCode code)
		{
			_code = code;
		}
		public ErrorCode Code
		{
			get
			{
				return _code;
			}
		}
		public override string Message
		{
			get
			{
				return _code.ToString();
			}
		}
	}
	public class Parser {

		virtual protected void StartElement(string name, ListDictionary attr)
		{
		}
		virtual protected void EndElement(string name)
		{
		}
		virtual protected void Characters(string text)
		{
		}

		enum State {
			Normal,
			TagNameFirst,
			TagNameTrail,
			Exclamation,
			EndTag,
			AttributeKeyFirst,
			AttributeKeyTrail,
			AttributeValue,
			AttributeValueQuoted,
			Comment0,
			Comment1,
			Comment2,
			CDATA0,
			CDATA1,
			CDATA2,
		}
		enum TagType {
			Question,
			Exclamation,
			StartEnd,
			Start,
			End,
		}
		bool isnamecharf(char c)
		{
			if (char.IsLetter(c)) return true;
			if (c == ':') return true;
			if (c == '_') return true;
			return false;
		}
		bool isnamechar(char c)
		{
			if (char.IsLetterOrDigit(c)) return true;
			if (c == ':') return true;
			if (c == '-') return true;
			if (c == '_') return true;
			if (c == '.') return true;
			return false;
		}
		State _state = State.Normal;
		TagType _tag_type = TagType.Question;
		string _tag_name;
		string _attr_key;
		char _quote;
		StringBuilder _entity = null;
		ListDictionary _attr = new ListDictionary();
		StringBuilder _sb = new StringBuilder();

		enum Encoding {
			UTF8,
			EUCJP,
			SHIFTJIS,
			ISO2022JP,
		}
		Encoding _encoding;

		void DispatchTag()
		{
			if (_tag_type == TagType.Question) {
				if (_tag_name == "xml") {
					foreach (DictionaryEntry e in _attr) {
						if (e.Key.ToString() == "encoding") {
							string str = e.Value.ToString();
							StringBuilder sb = new StringBuilder();
							for (int i = 0; i < str.Length; i++) {
								if (char.IsLetterOrDigit(str[i])) {
									sb.Append(char.ToUpper(str[i]));
								}
							}
							str = sb.ToString();
							if (str == "EUCJP") {
								_encoding = Encoding.EUCJP;
							} else if (str == "SHIFTJIS") {
								_encoding = Encoding.SHIFTJIS;
							} else if (str == "ISO2022JP") {
								_encoding = Encoding.ISO2022JP;
							}
						}
					}
				}
			} else {
				if (_tag_type == TagType.Start || _tag_type == TagType.StartEnd) {
					StartElement(_tag_name, _attr);
				}
				if (_tag_type == TagType.End || _tag_type == TagType.StartEnd) {
					EndElement(_tag_name);
				}
			}
		}

		protected void PutText(char c)
		{
			if (_entity == null) {
				if (c == '&') {
					_entity = new StringBuilder();
				} else {
					_sb.Append(c);
				}
			} else {
				if (c == ';') {
					string str = _entity.ToString();
					_entity = null;
					if (str.Length >= 2 && str[0] == '#') {
						int val = 0;
						try {
							if (str[1] == 'x' || str[1] == 'X') {
								val = System.Convert.ToInt32(str.Substring(2), 16);
							} else {
								val = System.Convert.ToInt32(str.Substring(1), 10);
							}
						} catch (System.Exception) {
						}
						_sb.Append((char)val);
					} else if (str == "amp") {
						_sb.Append('&');
					} else if (str == "lt") {
						_sb.Append('<');
					} else if (str == "gt") {
						_sb.Append('>');
					} else if (str == "quot") {
						_sb.Append('\"');
					} else if (str == "apos") {
						_sb.Append('\'');
					}
				} else {
					_entity.Append(c);
				}
			}
		}
		protected void Put(char c)
		{
			if (_state == State.Normal) {
				if (c == '<') {
					if (_sb.Length > 0) {
						Characters(_sb.ToString());
					}
					_sb = new StringBuilder();
					_attr.Clear();
					_state = State.TagNameFirst;
				} else {
					PutText(c);
				}
			} else if (_state == State.TagNameFirst) {
				if (isnamecharf(c)) {
					_sb.Append(c);
					_tag_type = TagType.Start;
					_state = State.TagNameTrail;
				} else if (c == '/') {
					_tag_type = TagType.End;
					_state = State.TagNameTrail;
				} else if (c == '?') {
					_tag_type = TagType.Question;
					_state = State.TagNameTrail;
				} else if (c == '!') {
					_tag_type = TagType.Exclamation;
					_state = State.Exclamation;
				} else {
					throw new MyException(ErrorCode.InvalidCharacter);
				}
			} else if (_state == State.TagNameTrail) {
				if (isnamechar(c)) {
					_sb.Append(c);
				} else {
					_tag_name = _sb.ToString();
					_sb = new StringBuilder();
					if (c == '>') {
						DispatchTag();
						_state = State.Normal;
					} else if (c == '/') {
						if (_tag_type != TagType.Start) {
							throw new MyException(ErrorCode.InvalidCharacter);
						}
						_tag_type = TagType.StartEnd;
						_state = State.EndTag;
					} else if (char.IsWhiteSpace(c)) {
						_state = State.AttributeKeyFirst;
					} else {
						throw new MyException(ErrorCode.InvalidCharacter);
					}
				}
			} else if (_state == State.AttributeKeyFirst) {
				if (isnamecharf(c)) {
					_sb.Append(c);
					_state = State.AttributeKeyTrail;
				} else if (c == '>') {
					DispatchTag();
					_state = State.Normal;
				} else if (c == '?') {
					if (_tag_type == TagType.Question) {
						_state = State.EndTag;
					} else {
						throw new MyException(ErrorCode.InvalidCharacter);
					}
				} else if (c == '/') {
					if (_tag_type != TagType.Start) {
						throw new MyException(ErrorCode.InvalidCharacter);
					}
					_tag_type = TagType.StartEnd;
					_state = State.EndTag;
				} else if (!char.IsWhiteSpace(c)) {
					throw new MyException(ErrorCode.InvalidCharacter);
				}
			} else if (_state == State.AttributeKeyTrail) {
				if (isnamechar(c)) {
					_sb.Append(c);
				} else if (c == '=') {
					_attr_key = _sb.ToString();
					_sb = new StringBuilder();
					_state = State.AttributeValue;
				} else {
					throw new MyException(ErrorCode.InvalidCharacter);
				}
			} else if (_state == State.AttributeValue) {
				if (c == '\'' || c == '\"') {
					_state = State.AttributeValueQuoted;
					_quote = c;
				} else {
					throw new MyException(ErrorCode.InvalidCharacter);
				}
			} else if (_state == State.AttributeValueQuoted) {
				if (c == _quote) {
					string val = _sb.ToString();
					_attr.Add(_attr_key, val);
					_sb = new StringBuilder();
					_state = State.AttributeKeyFirst;
				} else {
					PutText(c);
				}
			} else if (_state == State.EndTag) {
				if (c == '>') {
					DispatchTag();
					_state = State.Normal;
				} else {
					throw new MyException(ErrorCode.InvalidCharacter);
				}
			} else if (_state == State.Exclamation) {
				_sb.Append(c);
				if (_sb.Length == 2 && _sb.ToString() == "--") {
					_state = State.Comment0;
				} else if (_sb.Length == 7 && _sb.ToString() == "[CDATA[") {
					_state = State.CDATA0;
				}
			} else if (_state == State.Comment0) {
				if (c == '-') {
					_state = State.Comment1;
				} else {
					_state = State.Comment0;
				}
			} else if (_state == State.Comment1) {
				if (c == '-') {
					_state = State.Comment2;
				} else {
					_state = State.Comment0;
				}
			} else if (_state == State.Comment2) {
				if (c == '>') {
					_state = State.Normal;
				} else if (c == '-') {
					_state = State.Comment2;
				} else {
					_state = State.Comment0;
				}
			} else if (_state == State.CDATA0) {
				_sb.Append(c);
				if (c == ']') {
					_state = State.CDATA1;
				} else {
					_state = State.CDATA0;
				}
			} else if (_state == State.CDATA1) {
				_sb.Append(c);
				if (c == ']') {
					_state = State.CDATA2;
				} else {
					_state = State.CDATA0;
				}
			} else if (_state == State.CDATA2) {
				if (c == '>') {
					int n = _sb.Length;
					if (n > 9) {
						string text = _sb.ToString().Substring(7, n - 9);
						Characters(text);
					}
					_sb = new StringBuilder();
					_state = State.Normal;
				} else {
					_sb.Append(c);
					if (c == ']') {
						_state = State.CDATA2;
					} else {
						_state = State.CDATA0;
					}
				}
			}
		}

		int _ucs = 0;
		private void PutUTF8(byte b)
		{
			if ((b & 0x80) != 0) {
				if ((b & 0x40) != 0) {
					if ((b & 0x20) != 0) {
						if ((b & 0x10) != 0) {
							if ((b & 0x08) != 0) {
								if ((b & 0x04) != 0) {
									_ucs = 0x02 | (b & 0x01);
								} else {
									_ucs = 0x04 | (b & 0x03);
								}
							} else {
								_ucs = 0x08 | (b & 0x07);
							}
						} else {
							_ucs = 0x10 | (b & 0x0f);
						}
					} else {
						_ucs = 0x20 | (b & 0x1f);
					}
				} else {
					_ucs = (_ucs << 6) | (b & 0x3f);
					if ((_ucs & 0xfffff800) == 0x00000800) {
						Put((char)(_ucs & 0x000007ff));
						_ucs = 0;
					} else if ((_ucs & 0xffff0000) == 0x00010000) {
						Put((char)(_ucs & 0x0000ffff));
						_ucs = 0;
					} else if ((_ucs & 0xffe00000) == 0x00200000) {
						Put((char)(_ucs & 0x001fffff));
						_ucs = 0;
					} else if ((_ucs & 0xfc000000) == 0x04000000) {
						Put((char)(_ucs & 0x03ffffff));
						_ucs = 0;
					} else if ((_ucs & 0x80000000) == 0x80000000) {
						Put((char)(_ucs & 0x7fffffff));
						_ucs = 0;
					}
				}
			} else {
				Put((char)b);
			}
		}

		int _eucjp = 0;
		private void PutEUCJP(byte b)
		{
			if (b < 0x80) {
				Put((char)b);
			} else {
				if (_eucjp == 0) {
					_eucjp = b;
				} else {
					if (_eucjp == 0x8e) {
						Put(jcode.jistoucs(b));
					} else {
						ushort t = (ushort)(((_eucjp << 8) | b) & 0x7f7f);
						char x = jcode.jistoucs(t);
						Put(x);
					}
					_eucjp = 0;
				}
			}
		}

		int _sjis = 0;
		private void PutShiftJIS(byte b)
		{
			if (_sjis == 0) {
				if (b < 0x80) {
					Put((char)b);
				} else if ((b >= 0x81 && b <= 0x9f) || (b >= 0xe0 && b <= 0xfc)) {
					_sjis = b;
				} else if ((b >= 0xa1 && b <= 0xdf)) {
					Put(jcode.jistoucs(b));
				}
			} else {
				if ((b >= 0x40 && b <= 0x7e) || (b >= 0x80 && b <= 0xfc)) {
					ushort j = (ushort)((_sjis << 8) | b);
					j = jcode.jmstojis(j);
					char c = jcode.jistoucs(j);
					Put(c);
				}
				_sjis = 0;
			}
		}

		int _jis_mode = 0;
		int _jis_code = 0;
		private void PutISO2022JP(byte b)
		{
			if (b == 0x1b || b == 0x0f) {
				_jis_mode = b;
			} else if (b == 0x0e) {
				_jis_mode = 0;
			} else if (_jis_mode == 0x1b) {
				_jis_mode = (_jis_mode << 8) | b;
			} else if ((_jis_mode & 0x0000ff00) == 0x00001b00) {
				_jis_mode = (_jis_mode << 8) | b;
			} else {
				if (_jis_mode == 0x1b2849 || _jis_mode == 0x0f) {
					ushort j = (ushort)(b | 0x80);
					char c = jcode.jistoucs(j);
					Put(c);
				} else if (_jis_mode == 0x1b2440 || _jis_mode == 0x1b2442) {
					if (_jis_code == 0) {
						_jis_code = b;
					} else {
						ushort j = (ushort)((_jis_code << 8) | b);
						char c = jcode.jistoucs(j);
						Put(c);
						_jis_code = 0;
					}
				} else {
					Put((char)b);
				}
			}
		}

		protected void Parse(char[] chars, int len)
		{
			for (int i = 0; i < len; i++) {
				Put(chars[i]);
			}
		}
		protected void Parse(byte[] bytes, int len)
		{
			for (int i = 0; i < len; i++) {
				switch (_encoding) {
				case Encoding.EUCJP:
					PutEUCJP(bytes[i]);
					break;
				case Encoding.SHIFTJIS:
					PutShiftJIS(bytes[i]);
					break;
				case Encoding.ISO2022JP:
					PutISO2022JP(bytes[i]);
					break;
				default:
					PutUTF8(bytes[i]);
					break;
				}
			}
		}
		protected void Parse(Stream stream)
		{
			byte[] bytes = new byte[4096];
			while (true) {
				int n = stream.Read(bytes, 0, bytes.Length);
				if (n == 0) {
					break;
				}
				Parse(bytes, n);
			}
		}
	}
}
